# protocolrunner.py - script functions developed for integration with ScienceCloud via PipelinePilot server
# Use of this code is subject to the terms of the StarDrop License Agreement and use, copying or redistribution of this file for use by other persons or organisations is prohibited
# Copyright(C) Optibrium Ltd 2017
# stardrop-support@optibrium.com
#

"""
Module handles running PipelinePilpot protocols to access ScienceCloud data.
"""

import urllib.request as request
import urllib.parse as parse
import json
import time
import threading

import sciencecloudconfig as config

def _checkResults(job_id):
    """Check the progress of a submitted job.
    @param: job_id: The id for the job.
    @return: A string describing status.
    """
    url_check = config.base_url + '/jobs/%s/status?%s' % (job_id,parse.urlencode({'_format':'json'}))
    try:
        query = request.Request(url_check)
        response = request.urlopen(query)
        return json.loads(response.read()).get("data",{}).get("value","Unknown")
    except request.HTTPError as he:
        return "error"
        
           
def _getJobResults(job_id):
    """Retrieve the results of a completed job.
    @param: job_id: The id for the job.
    @return: The raw data from the protocol.
    """
    url_retrieve = config.base_url + '/jobs/%s/results/all?%s' % (job_id,parse.urlencode({'_format':'json'}))
    try:
        query = request.Request(url_retrieve)
        response = request.urlopen(query)
        data = json.loads(response.read())
        return data
    except request.HTTPError as he:
        print(he.read())
        
def _getJobError(job_id):
    """Retrieve the error for a failed job.
    @param: job_id: The id for the job.
    @return: The raw data from the protocol.
    """
    url_retrieve_error = config.base_url + '/jobs/%s/error?%s' % (job_id,parse.urlencode({'_format':'json'}))
    try:
        query = request.Request(url_retrieve_error)
        response = request.urlopen(query)
        data = json.loads(response.read())
        print(data)
        return data.get("ErrorStack","Unknown error")
    except request.HTTPError as he:
        print(he.read())
        
def runProtocol(protocol, credentials,parameters,stop): 
    """Run a protocol.
    @param: protocol: The name of the protocol to run.
    @param: credentials: a CredentialsManager object.
    @param: parameters: Additional parameters to pass to the protocol, as a dictionary.
    @param: stop: A monitor to check if user has cancelled operation.
    @return: Dictionary containing the raw data from the protocol, or an error.
    """
    if not credentials:
        return {"error":"No credentials supplied"}
    url = config.base_url + "/auth/launchjob?"
    data = {}
    data["SCUser"] = credentials.getUser()
    data["SCPassword"] = credentials.getPassword()
    data["_protocol"] = config.protocol_folder + protocol
    data["_streamdata"] = "*"
    data["_format"] = "json"
    data["_blocking"] = "False"
    for key in parameters:
        data[key] = parameters[key]
    params = parse.urlencode(data) 
    url += params
    try:
        query = request.Request(url)
        response = request.urlopen(query)
        job_ID = json.loads(response.read()).get("data",{}).get("value","Unknown")
        complete = False
        while not complete:
            if stop.set():
                print("Thread stopping")
                break
            status = _checkResults(job_ID)
            if status.lower() in ("complete","error"):
                complete = True
            else:
                time.sleep(5)
        if status.lower() == "complete":
            return _getJobResults(job_ID)
        error = _getJobError(job_ID)
        return {"error":"Failed to run protocol: %s\n%s" % (protocol,error)}
    except request.HTTPError as he:
        if he.code == 401:#authentication error
            credentials.clear()
            return {"error": "Authentication error, please check user name and password"}
        try:
            print(he.read())
        except:
            print(he)
        return {"error": "HTTP error, see console for details"}
    except request.URLError as ue:
        print(ue)
        return {"error": "URL error, see console for details"}
    except Exception as e:
        print(type(e))
        print(e)
        return {"error": "Unexpected error, see console for details"}
        
class ThreadMonitor:
    """Simple monitor class to give calling code a way to stop thread running when user cancels the operation.
    """
    def __init__(self):
        """Initialise a ThreadMonitor."""
        self._set = False
        self._lock = threading.Lock()
        
    def set(self):
        """Check the flag on the monitor to tell te thread to stop.
        @return: True if stop has been called, False otherwise
        """
        return self._set
        
    def stop(self):
        """Set the flag on the monitor to tell te thread to stop."""
        self._lock.acquire()
        self._set = True
        self._lock.release()
        
class ProtocolRunnerThread(threading.Thread):
    """Thread class to handle running a protocol in the background."""
    def __init__(self,protocol,credentials,parameters = {}):
        """Initialise a thread object.
        @param: protocol: The name of the protocol to run.
        @param: credentials: a CredentialsManager object.
        @param: parameters: Additional parameters to pass to the protocol, as a dictionary.
        """
        threading.Thread.__init__(self)
        self._protocol = protocol
        self._credentials = credentials
        self._parameters = parameters
        self._data = {}
        self._monitor = ThreadMonitor()
        self.daemon = True
        
    def run(self):
        """Run the thread."""
        self._data = runProtocol(self._protocol, self._credentials, self._parameters, self._monitor)
        
    def protocolResult(self):
        """Get the results of the protocol.
        @return: R=The raw data returned from the protocol.
        """
        return self._data
        
    def stop(self):
        """Set a flag to stop the thread."""
        self._monitor.stop()
    
